#include "packet.h"

enum FD_STATE read_packet_from_tcp(int fd, struct buffer_packet* bp) {
  ssize_t nread;
  size_t pkt_size_size = sizeof(bp->ip.ap.str.size);
  if (bp->mode != BP_READING) return FDS_ERR;

  while (bp->aread < pkt_size_size) {
    nread = read(fd, &(bp->ip.ap.raw) + bp->aread, pkt_size_size - bp->aread);
    if (nread == 0) return FDS_AGAIN;
    if (nread == -1 && errno == EAGAIN) return FDS_AGAIN;
    if (nread == -1) return FDS_ERR;
    bp->aread += nread;
  }

  while (bp->aread < bp->ip.ap.str.size) {
    nread = read(fd, &(bp->ip.ap.raw) + bp->aread, bp->ip.ap.str.size - bp->aread);
    if (nread == 0) return FDS_AGAIN;
    if (nread == -1 && errno == EAGAIN) return FDS_AGAIN;
    if (nread == -1) return FDS_ERR;
    bp->aread += nread;
  }

  bp->mode = BP_WRITING;
  bp->awrite = 0;

  return FDS_READY;
}

enum FD_STATE write_packet_to_tcp(int fd, struct buffer_packet* bp) {
  ssize_t nwrite;

  if (bp->mode != BP_WRITING) return FDS_ERR;
  while (bp->awrite < bp->ip.ap.str.size) {
    nwrite = send(fd, &(bp->ip.ap.raw), bp->ip.ap.str.size, 0);
    if (nwrite == -1 && errno == EAGAIN) return FDS_AGAIN;
    if (nwrite == -1) return FDS_ERR;
    bp->awrite += nwrite;
  }

  bp->mode = BP_READING;
  bp->aread = 0;

  return FDS_READY;
}

enum FD_STATE write_packet_to_udp(int fd, struct buffer_packet* bp) {
  ssize_t nwrite;
  size_t bytes_to_send;
  size_t pkt_header_size = sizeof(bp->ip.ap.str) - sizeof(char);
  struct sockaddr_in addr = {
    .sin_family = AF_INET,
    .sin_port = htons(bp->ip.ap.str.port),
  };
  inet_aton("127.0.0.1", &addr.sin_addr);
  socklen_t addr_len = sizeof(addr);

  if (bp->mode != BP_WRITING) return FDS_ERR;

  bytes_to_send = bp->ip.ap.str.size - pkt_header_size;
  nwrite = sendto(fd,
                  &(bp->ip.ap.str.payload),
                  bytes_to_send,
                  0,
                  (struct sockaddr*)&addr,
                  addr_len);

  if (nwrite == -1 && errno == EAGAIN) return FDS_AGAIN;
  if (nwrite != bytes_to_send) return FDS_ERR;

  bp->mode = BP_READING;
  bp->aread = 0;

  return FDS_READY;
}

enum FD_STATE read_packet_from_udp (int fd, struct buffer_packet* bp) {
  ssize_t nread;
  if (bp->mode != BP_READING) return FDS_ERR;

  struct sockaddr_in addr = {0};
  socklen_t addr_len = sizeof(addr);
  size_t pkt_header_size = sizeof(bp->ip.ap.str) - sizeof(char); // We remove the payload
  size_t udp_packet_size = sizeof(struct internet_packet) - pkt_header_size;

  nread = recvfrom(fd,
                   &(bp->ip.ap.str.payload),
                   udp_packet_size,
                   MSG_TRUNC,
                   (struct sockaddr*)&addr,
                   &addr_len);

  if ((int)nread > (int)udp_packet_size) return FDS_ERR;
  if (nread == -1 && errno == EAGAIN) return FDS_AGAIN;
  if (nread == -1) return FDS_ERR;

  bp->ip.ap.str.port = ntohs(addr.sin_port);
  bp->ip.ap.str.size = nread + pkt_header_size;

  bp->mode = BP_WRITING;
  bp->awrite = 0;

  return FDS_READY;
}

